geninv <- function(nsam,a,b) {

    job=11
    xcall = ssvdc(a,nsam,nsam,s,e,u,v,work,job)
    s=xcall$s; e=xcall$e; u=xcall$u; v=xcall$v
    
    if (xcall$info != 0) {
      cat('*********************************************\n')
      cat(' error in matrix inverse function\n')
      cat('*********************************************\n')
    }
    #  set up the diagonal matrix of eigenvalues
    i=which(s>.000001); if (length[i]>0) eig[i,i]=1/s[i]

    #   put them back together, transpose exp, u, mult u'exp'v      

    u=y %*% eig %*% t(v)
    return(t(u))
}

ssvdc <- function(x,n,p,work,job,info) {

    #     ssvdc is a subroutine to reduce a real nxp matrix x by orthogonal transformations u and v to diagonal form.  the
    #     diagonal elements s[i] are the singular values of x.  the columns of u are the corresponding left singular vectors,
    #     and the columns of v the right singular vectors.

    #     on entry
    #
    #         x         real(ldx,p), where ldx >= n.  x contains the matrix whose singular value
    #                   decomposition is to be computed.  x is   destroyed by ssvdc.
    #
    #         ldx       integer.             ldx is the leading dimension of the array x.
    #
    #         n         integer.             n is the number of columns of the matrix x.
    #
    #         p         integer.             p is the number of rows of the matrix x.
    #
    #         ldu       integer.             ldu is the leading dimension of the array u. (see below).
    #
    #         ldv       integer.             ldv is the leading dimension of the array v. (see below).
    #
    #         work      real(n).             work is a scratch array.
    #
    #         job       integer.             job controls the computation of the singular  vectors.  it has the decimal expansion ab
    #                   with the following meaning
    #
    #                        a == 0    do not compute the left singular vectors.
    #                        a == 1    return the n left singular vectors in u.
    #                        a >= 2    return the first min(n,p) singular vectors in u.
    #                        b == 0    do not compute the right singular vectors.
    #                        b == 1    return the right singular vectors in v.
    #
    #     on return
    #
    #         s         real(mm), where mm=min(n+1,p). the first min(n,p) entries of s contain the
    #                   singular values of x arranged in descending order of magnitude.
    #
    #         e         real(p).           e ordinarily contains zeros.  however see the discussion of info for exceptions.
    #
    #         u         real(ldu,k), where ldu >= n.  if joba == 1 {  k == n,
    #                                                 if joba >= 2 {  k == min(n,p).
    #                   u contains the matrix of right singular vectors.
    #                   u is not referenced if joba == 0.  if n <= p  or if joba == 2, { u may be identified with x in the subroutine call.
    #
    #         v         real(ldv,p), where ldv >= p.  v contains the matrix of right singular vectors.
    #                   v is not referenced if job == 0.  if p <= n, { v may be identified with x in the subroutine call.
    #
    #         info      integer.
    #                   the singular values (and their corresponding singular vectors) s(info+1),s(info+2),...,s[m]
    #                   are correct (here m=min(n,p)).  thus if info == 0, all the singular values and their vectors are correct.
    #                   in any event, the matrix  b = trans(u)*x*v is the bidiagonal matrix with the elements of s on its diagonal and the
    #                   elements of e on its super-diagonal (trans(u)  is the transpose of u).  thus the singular values of x and b are the same.
    #
    #     linpack. this version dated 03/19/79 .  g.w. stewart, university of maryland, argonne national lab.
    #
    #     ***** uses the following functions and subprograms.
    #
    #     external srot
    #     blas saxpy,sdot,sscal,sswap,snrm2,srotg
    #     fortran abs,amax1,max,min,mod,sqrt
    #

    maxit = 30  #     set the maximum number of iterations.
    wantu = wantv = FALSE   #     determine what is to be computed.
    
    jobu = (job %% 100)/10
    ncu = n
    if (jobu  >  1) ncu = min(n,p)
    if (jobu  !=  0) wantu = TRUE
    if (mod(job,10)  !=  0) wantv = TRUE

    #     reduce x to bidiagonal form, storing the diagonal elements  in s and the super-diagonal elements in e.

    info = 0; nct = min(n-1,p); nrt = max(0,min(p-2,n)); lu = max(nct,nrt)
    if (lu  >=  1) {
    for (l in 1:lu) {
       lp1 = l + 1
       if (l  <=  nct) {
          s[l] = snrm2(n-l+1,x[l,l],1)  # compute the transformation for the l-th column and place the l-th diagonal in s[l].
          if (s[l]  !=  0) {
             if (x[l,l]  !=  0) s[l] = sign(s[l],x[l,l])
             xll = sscal(n-l+1,1/s[l],x[l,l],1)
             x[l,l] = 1 + xll
          }
          s[l] = -s[l]
       }
       if (p  >=  lp1) {
          for (j in  lp1:p) {
             if (l  <=  nct) {
                if (s[l]  !=  0) {
                   t = -sdot(n-l+1,x[l,l],1,x[l,j],1)/x[l,l]  # apply the transformation.
                   x[l,j] = saxpy(n-l+1,t,x[l,l],1,x[l,j],1)
                }
             }
             e[j] = x[l,j]  # place the l-th row of x into  e for the  sub == ent calculation of the row transformation.
          }
       }
       if (wantu  &&  l  <=  nct) {
           u[l:n,l] = x[l:n,l]  # place the transformation in u for sub == ent back multiplication.       
       }
       if (l  <=  nrt) {
          e[l] = snrm2(p-l,e[lp1],1)  # compute the l-th row transformation and place the l-th super-diagonal in e[l].
          if (e[l]  !=  0) {
             if (e[lp1]  <  0) e[l] = -e[l]  #sign(e[l],e[lp1])
             elp1 = sscal(p-l,1/e[l],e[lp1],1)
             e[lp1] = 1 + elp1
          }
          e[l] = -e[l]
          if (lp1  <=  n  &&  e[l]  !=  0) {
             work[lp1:n] = 0#    apply the transformation.

             for (j in  lp1:p) {
                work[lp1] = saxpy(n-l,e[j],x(lp1,j),1,work[lp1],1)
             }
             for (j in  lp1:p) {
                x[lp1,j] = saxpy(n-l,-e[j]/e[lp1],work[lp1],1,x[lp1,j],1)
             }
          }
          if (wantv) v[lp1:p,l] = e[lp1:p]  #   place the transformation in v for sub == ent back multiplication.
       }
    }
    }

    m = min(p,n+1)  #    set up the final bidiagonal matrix or order m.
    nctp1 = nct + 1; nrtp1 = nrt + 1
    if (nct  <  p) s[nctp1] = x[nctp1,nctp1]
    if (n  <  m) s[m] = 0
    if (nrtp1  <  m) e[nrtp1] = x[nrtp1,m]
    e[m] = 0

    if (wantu) {  #    if  == ired, generate u.
       if (ncu  >=  nctp1) {
          for (j in  nctp1:ncu) {
             u[1:n,j] = 0;  u[j,j] = 1
          }
       }
       if (nct  >=  1) {
          for (ll in 1:nct) {
             l = nct - ll + 1
             if (s[l]  !=  0) {
                lp1 = l + 1
                if (ncu  >=  lp1) {
                   for (j in  lp1:ncu) {
                     t = -sdot(n-l+1,u[l,l],1,u[l,j],1)/u[l,l]
                     u[l,j] = saxpy(n-l+1,t,u[l,l],1,u[l,j],1)
                   }
                }
                u[l,l] = 1 + sscal(n-l+1,-1,u[l,l],1)
                if (lm1  >=  1) u[1:(l-1),l] = 0
             }
             else
                   u[1:n,l] = 0
             u[l,l] = 1
          }
       }
    }
    }

    if (wantv) {  #  if it is  == ired, generate v.
       for (ll in 1:p) {
          l = p - ll + 1; lp1 = l + 1
          if (l  <=  nrt) {
             if (e[l]  !=  0) {
                for (j in  lp1:p) {
                   t = -sdot(p-l,v[lp1,l],1,v(lp1,j),1)/v[lp1,l]
                   call saxpy(p-l,t,v[lp1,l],1,v(lp1,j),1)
                }
             }
          }
          for (i in  1, p
             v[i,l] = 0
          }
          v[l,l] = 1
       }
    }

    #     main iteration loop for the singular values.

    mm = m; iter = 0
  #360 continue
    while (1) {

      #        quit if all the singular values have been found.

      #     ...exit
         if (m  ==  0) return(list(info=NULL))

      #        if too many iterations have been performed, set flag and return.

         if (iter  >=  maxit) return(list(info=m))


      #        this section of the program inspects for negligible elements in the s and e arrays. on completion the variables kase and l are set as follows.

      #           kase = 1     if s[m] and e(l-1) are negligible and l < m
      #           kase = 2     if s[l] is negligible and l < m
      #           kase = 3     if e(l-1) is negligible, l < m, and s[l], ..., s[m] are not negligible (qr step).
      #           kase = 4     if e(m-1) is negligible (convergence).

         for (ll in 1:m) {
            l = m - ll
            if (l  ==  0) stop('geninv - exit')
            test = abs(s[l]) + abs(s(l+1))
            ztest = test + abs(e[l])
            if (ztest  ==  test) {
               e[l] = 0
               stop('geninv - exit2')
            }
         }

         if (l  ==  (m - 1)) {
            kase = 4
         else
            lp1 = l + 1
            mp1 = m + 1
            do lls = lp1:mp1
               ls = m - lls + lp1
               if (ls  ==  l) exit
               test = 0
               if (ls  !=  m) test = test + abs(e(ls))
               if (ls  !=  l + 1) test = test + abs(e(ls-1))
               ztest = test + abs(s(ls))
               if (ztest  ==  test) {
                  s(ls) = 0
                  exit
               }
            }
            if (ls  ==  l) {
               kase = 3
            else
               if (ls  ==  m) {
                  kase = 1
               else
                  kase = 2
                  l = ls
               }
            }
         }
         l = l + 1

      #        perform the task indicated by kase.

        # go to (490,520,540,570), kase
        if (kase==1) {
  490       mm1 = m - 1            #   deflate negligible s[m].
            f = e(m-1)
            e(m-1) = 0
            do kk = l, mm1
               k = mm1 - kk + l
               t1 = s[k]
               call srotg(t1,f,cs,sn)
               s[k] = t1
               if (k  !=  l) {
                  f = -sn*e(k-1)
                  e(k-1) = cs*e(k-1)
               }
               if (wantv) call srot(p,v(1,k),1,v(1,m),1,cs,sn)
            }
        }
        if (kase==2) {
            f = e(l-1)  #    split at negligible s[l].
            e(l-1) = 0
            for (k in  l:m) {
               t1 = s[k]
               call srotg(t1,f,cs,sn)
               s[k] = t1
               f = -sn*e[k]
               e[k] = cs*e[k]
               if (wantu) call srot(n,u(1,k),1,u(1,l-1),1,cs,sn)
            }
        }
        if (kase==3) {
            scale = amax1(abs(s[m]),abs(s(m-1)),abs(e(m-1)),abs(s[l]),abs(e[l]))  #    perform one qr step.  calculate the shift.
            sm = s[m]/scale
            smm1 = s(m-1)/scale
            emm1 = e(m-1)/scale
            sl = s[l]/scale
            el = e[l]/scale
            b = ((smm1 + sm)*(smm1 - sm) + emm1**2)/2
            c = (sm*emm1)**2
            shift = 0
            if (b  !=  0  ||  c  !=  0) {
               shift = sqrt(b**2+c)
               if (b  <  0) shift = -shift
               shift = c/(b + shift)
            }
            f = (sl + sm)*(sl - sm) + shift
            g = sl*el

            mm1 = m - 1  #   chase zeros.
            for (k in  l, mm1
               call srotg(f,g,cs,sn)
               if (k  !=  l) e(k-1) = f
               f = cs*s[k] + sn*e[k]
               e[k] = cs*e[k] - sn*s[k]
               g = sn*s(k+1)
               s(k+1) = cs*s(k+1)
               if (wantv) call srot(p,v(1,k),1,v(1,k+1),1,cs,sn)
               call srotg(f,g,cs,sn)
               s[k] = f
               f = cs*e[k] + sn*s(k+1)
               s(k+1) = -sn*e[k] + cs*s(k+1)
               g = sn*e(k+1)
               e(k+1) = cs*e(k+1)
               if (wantu  &&  k  <  n) call srot(n,u(1,k),1,u(1,k+1),1,cs,sn)
            }
            e(m-1) = f
            iter = iter + 1
        }
        if (kase==4) {
  570    continue  #    convergence.      make the singular value  positive.

            if (s[l]  <  0) {
               s[l] = -s[l]
               if (wantv) call sscal(p,-1,v[1,l],1)
            }

      #           order the singular value.
            do while (l  !=  mm) {
                if (s[l]  >=  s(l+1)) exit
                t = s[l]
                s[l] = s(l+1)
                s(l+1) = t
                if (wantv  &&  l  <  p) {
                    xcall = sswap(p,v[1,l],1,v[1,l+1],1); v[1,l]=xcall$sx; v[1,l+1]=xcall$sy
                }
                if (wantu  &&  l  <  n) {
                    call sswap(n,u[1,l],1,u[1,l+1],1); u[1,l]=xcall$sx; u[1,l+1]=xcall$sy
                l = l + 1
            }
            iter = 0; m = m - 1
        }
    }
    return
}

snrm2 <- function snrm2(n, sx, incx) {

    #   small is used to prevent values below small from being returned. see the end of the program for usage.
    real small
    small = 8.43e-37
    zero=0; one=1

    #     euclidean norm of the n-vector stored in sx() with storage increment incx .
    #     if    n  <=  0 return with result = 0.
    #     if n  >=  1 { incx must be  >=  1

    #           c.l.lawson, 1978 jan 08

    #     four phase method     using two built-in constants that are  hopefully applicable to all machines.
    #         cutlo = maximum of  sqrt(u/eps)  over all known machines.
    #         cuthi = minimum of  sqrt(v)      over all known machines.
    #     where
    #         eps = smallest no. such that eps + 1.  >  1.
    #         u   = smallest positive no.   (underflow limit)
    #         v   = largest  no.            (overflow  limit)
    #
    #     brief outline of algorithm..
    #
    #     phase 1    scans zero components.
    #     move to phase 2 when a component is nonzero and  <=  cutlo
    #     move to phase 3 when a component is  >  cutlo
    #     move to phase 4 when a component is  >=  cuthi/m
    #     where m = n for x() real and m = 2*n for complex.
    #
    #     values for cutlo and cuthi..
    #     from the environmental parameters listed in the imsl converter document the limiting values are as follows..
    #     cutlo, s.p.   u/eps = 2**(-102) for  honeywell.  close seconds are univac and dec at 2**(-103) thus cutlo = 2**(-51) = 4.44089e-16
    #     cuthi, s.p.   v = 2**127 for univac, honeywell, and dec. thus cuthi = 2**(63.5) = 1.30438e19
    #     cutlo, d.p.   u/eps = 2**(-67) for honeywell and dec.   thus cutlo = 2**(-33.5) = 8.23181d-11
    #     cuthi, d.p.   same as s.p.  cuthi = 1.30438d19
    #     data cutlo, cuthi / 8.232d-11,  1.304d19 /
    #     data cutlo, cuthi / 4.441e-16,  1.304e19 /
    data cutlo, cuthi / 4.441e-16,  1.304e19 /

    if(n  >  0) go to 10
       snrm2  = zero
       go to 300

    #   10 assign 30 to next
  10  next=30
      sum = zero; nn = n * incx
      #                                  gin main loop
      i = 1
      # 20  go to next,(30, 50, 70, 110)
  20  if(next == 50) goto 50
      if(next == 70) goto 70
      if(next == 110) goto 110
      if( abs(sx[i])  >  cutlo) go to 85
      next=50
      xmax = zero
      #                        phase 1.  sum is zero

   50 if( sx[i]  ==  zero) go to 200
      if( abs(sx[i])  >  cutlo) go to 85

      #                                prepare for phase 2.
      next=70
      go to 105
      #                                prepare for phase 4.
  100 i = j
      next=110
      sum = (sum / sx[i]) / sx[i]
  105 xmax = abs(sx[i])
      go to 115

      #                   phase 2.  sum is small.  scale to avoid destructive underflow.

   70 if( abs(sx[i])  >  cutlo ) go to 75

      #                     common code for phases 2 and 4. in phase 4 sum is large.  scale to avoid overflow.

  110 if( abs(sx[i])  <=  xmax ) go to 115
         sum = one + sum * (xmax / sx[i])**2
         xmax = abs(sx[i])
         go to 200

  115 sum = sum + (sx[i]/xmax)**2
      go to 200

      #                  prepare for phase 3.

   75 sum = (sum * xmax) * xmax

      #     for real or d.p. set hitest = cuthi/n
      #     for complex      set hitest = cuthi/(2*n)

   85 hitest = cuthi/float( n )

      #                   phase 3.  sum is mid-range.  no scaling.

      do 95 j =i,nn,incx
         if(abs(sx[j])  >=  hitest) go to 100
         sum = sum + sx[j]**2
   95 continue
      snrm2 = sqrt( sum )
      go to 300

  200 continue
      i = i + incx
      if ( i  <=  nn ) go to 20

      #              end of main loop.
      snrm2 = xmax * sqrt(sum)  #   compute square root and adjust for scaling.
  300 continue

      # *** 6 jan 84, j. fried : 
      #     this next line of fortran was added to prevent values below ms-fortran's specified minimum from being returned.  such 
      #     numbers have appeared during testing of the sqrdc routine (test #9 in sq).  the value of small was taken from the 
      #     ms-fortran manual.  remember, this only applies to pc/ms-fortran.

      if ( snrm2  <=  small ) snrm2 = small
      return
}

sscal <- function(n,da,dx,incx) {

      #     scales a vector by a constant. uses unrolled loops for increment == al to one.  jack dongarra, linpack, 3/11/78.
    dx=0
    if(n <= 0)return
    if(incx != 1) {
      nincx = n*incx  #         code for increment not == al to 1
      dx[seq(1,nincx,incx)] = da*dx[seq(1,nincx,incx)]
      return(dx)
    }
      #        code for increment == al to 1.       clean-up loop

    m = mod(n,5)
    if( m  !=  0 ) {
      dx[1:m] = da*dx[1:m]
      if( n  <  5 ) return
    }
    mp1 = m + 1
    for (i in seq(mp1,n,5)) dx[i+0:4] = da*dx[i+0:4]
    return(dx)
}

sdot <- function(n,sx,incx,sy,incy) { # forms the dot product of two vectors.  uses unrolled loops for increments == al to one.  jack dongarra, linpack, 3/11/78.
    sdot = temp = 0
    if(n <= 0)return
    if(!(incx == 1 && incy == 1)) {
        #        code for u == al increments or == al increments not == al to 1
        ix = iy = 1
        if(incx < 0)ix = (-n+1)*incx + 1
        if(incy < 0)iy = (-n+1)*incy + 1
        for (i in 1:n) { temp = temp + sx[ix]*sy[iy]; ix = ix + incx; iy = iy + incy }
        sdot = temp
        return
    }
    m = (n %%,5)  #         code for both increments == al to 1.    clean-up loop
    if( m  !=  0 ) {
        temp = sum(sx[1:m]*sy[1:m])
        if( n  <  5 ) return(temp)
    }
    mp1 = m + 1
    i=seq(mp1,n,5)
    temp = sum(sx[i]*sy[i]+sx(i+1)*sy[i+1)+sx(i+2)*sy[i+2)+sx(i+3)*sy[i+3)+sx(i+4)*sy[i+4)
    return(temp)
}

saxpy <- function(n,sa,sx,incx,sy,incy) {  #  constant times a vector plus a vector.  uses unrolled loops for increments == al to one.
      #                                                   jack dongarra, linpack, 3/11/78.
    sy=0
    if(n <= 0)return(sy)
    if (sa  ==  0) return(sy)
    if(!(incx == 1 && incy == 1)) {

        #        code for u == al increments or == al increments not == al to 1
        ix = iy = 1
        if(incx < 0)ix = (-n+1)*incx + 1
        if(incy < 0)iy = (-n+1)*incy + 1
        for (i in 1:n) {
            sy[iy] = sy[iy] + sa*sx[ix]
            ix = ix + incx; iy = iy + incy
        }
        return(sy)
    }
      #        code for both increments == al to 1.    clean-up loop
    m = (n %% 4)
    if( m  !=  0 ) {
        sy[1:m] = sy[1:m] + sa*sx[1:m]
        if( n  <  4 ) return(sy)
    }
    mp1 = m + 1
    for (i in seq(mp1,n,4)) 
        sy[i+0:3] = sy[i+0:3] + sa*sx[i+0:3]
    return(sy)
}

srotg <- function(sa,sb,c,s) { #    construct givens plane rotation.  jack dongarra, linpack, 3/11/78.

    roe = sb
    if( abs(sa)  >  abs(sb) ) roe = sa
    scale = abs(sa) + abs(sb)
    if( scale  ==  0) { c = 1; s = r = 0}  else 
      r = scale*sqrt((sa/scale)^2 + (sb/scale)^2)
      if (roe<0) r=-r  # r = sign(1.0,roe)*r
      c = sa/r; s = sb/r
    }
    z = 1
    if( abs(sa)  >  abs(sb) ) z = s
    if( abs(sb)  >=  abs(sa)  &&  c  !=  0 ) z = 1/c
    return(list(sa=r, sb=z))
}

srot <- function(n,sx,incx,sy,incy,c,s) { #    applies a plane rotation. jack dongarra, linpack, 3/11/78.

    if(n <= 0)return
    if(!(incx == 1 && incy == 1)) {

        #       code for u == al increments or == al increments not == al to 1
        ix = iy = 1
        if(incx < 0)ix = (-n+1)*incx + 1
        if(incy < 0)iy = (-n+1)*incy + 1
        for (i in  1,n
          stemp = c*sx(ix] + s*sy[iy]
          sy[iy] = c*sy[iy] - s*sx[ix]
          sx[ix] = stemp; ix = ix + incx; iy = iy + incy
        }
        return(list(sx=sx, sy=sy))

      #       code for both increments == al to 1
    }
    for (i in  1:n) {
        stemp = c*sx[i] + s*sy[i]
        sy[i] = c*sy[i] - s*sx[i]
        sx[i] = stemp
    }
    return(list(sx=sx, sy=sy))
}

sswap <- (n,sx,incx,sy,incy) {

      #     interchanges two vectors. uses unrolled loops for increments == al to 1.  jack dongarra, linpack, 3/11/78.

    if(n <= 0)return
    if(!(incx == 1 && incy == 1)) {

        #       code for u == al increments or == al increments not == al to 1

        ix = iy = 1
        if(incx < 0)ix = (-n+1)*incx + 1
        if(incy < 0)iy = (-n+1)*incy + 1
        for (i in  1:n) {
          stemp = sx[ix]
          sx[ix] = sy[iy]
          sy[iy] = stemp
          ix = ix + incx
          iy = iy + incy
        }
        return(list(sx=sx, sy=sy))
    }
    #       code for both increments == al to 1
    #       clean-up loop

    m = (n %% 3)
    if( m  !=  0 ) { 
        stemp = sx; sx = sy; sy = stemp
        if( n  <  3 ) return
    }
    mp1 = m + 1
    for (i in seq(mp1,n,3)) {
      stemp = sx[i+0:2]
      sx[i+0:2] = sy[i+0:2]
      sy[i+0:2] = stemp
      
    }
    return
}

