xspac <- function(a,b,n) return(a + (1:n)*(1/(n+1)*(b-a)))

loess <- function(x,y,nx,lcnt,yrs,yrf,yrl,oof2) {

    #   This subroutine computes  a LOESS smooth of variable y as a function of x.
    #
    #    Usage: call loess(fx,fy,x,y,nx)

    #  Here, fx consists of OONP (global, default = 100) points evenly spaced from the minimum x to the maximum x, and fy consists of the
    #  fitted values at those points.  Other globals are OOF (tension parameter, default = 0.5), OOSHOW and OOF ==  the latter two
    #  controlling whether and how f == ently the procs report progress. A final global OOROBUST determines if robust smoothing is to be
    #  used(default = 1).  The variable nx is the number of points in the vector x.
    #
    #  Local smoothing is linear with neighborhood weights determined by the bi-squarefunction.
    #

    #common /loes/oof,oonp,oof == ooshow,oorbst

    oonp=100; oof=oof2; oorbst=ooshow=0
    fclose=max(floor(nx*oof),1)
    iyrs=floor(yrl-yrf+1.0)
    rhos=rep(1,nx)

    if(oorbst == 1){
      alleps=y
      for (i in 1:nx) alleps[i]=y[i]-fitted(x,y,rhos,fclose,x[i],nx)
      tmp=median(abs(alleps,nx)*6.0)
      rhos=bisq(alleps/tmp,nx)
      for (i in 1:nx) alleps[i]=y[i]-fitted(x,y,rhos,fclose,x[i],nx)
      tmp=median(abs(alleps,nx)*6.0)
      rhos=bisq(alleps/tmp,nx)
    }
    xmin=x[1]; xmaxi=x[nx]; xs=allfit=xspac(xmin,xmaxi,oonp); 
    for (i in 1:length(yrs)) {
      if(yrs[i] >= xmin && yrs[i] <= xmaxi) lcnt[i]=fitted(x,y,rhos,fclose,yrs[i],nx)
    }
    for (i in 1:oonp) allfit[i]=fitted(x,y,rhos,fclose,xs[i],oonp)
    return(list(fx=xs, fy=allfit, lcnt=lcnt))
}

fitted <- function(x,y,rhos,fclose,t,n){
    tlamda=lmda(x[1:n],fclose,t,n); v=bisq((x-t)/tlamda,n); a=b=c=d=e=0
    for (i in 1:n) {
        v[i]=(v[i]+1e-50)*rhos[i]
        a=a+v[i]*y[i]
        b=b+v[i]
        c=c+v[i]*x[i]
        d=d+v[i]*x[i]*y[i]
        e=e+v[i]*x[i]*x[i]
    }
    b1=(b*d-c*a)/(b*e-c*c)
    b0=(a-c*b1)/b
    return(b0+b1*t)
}

bisq <- function(x,n){  #    compute (1-x^2)^2 for abs(x[i]<1
    rslt=rep(0,n)
    i=which(abs(x)<1)
    rslt[i]=(1-x[i]*x[i])*(1-x[i]*x[i])
    return(rslt)
}

lmda <- function(x,fclose,t,n){
    ax=abs(x-t); o=order(ax); return(ax[o[fclose]])
}
