      subroutine loess(xs,allfit,x,y,nx,LCNT,YRS,yrf,yrl,oof2)

c             This subroutine computes  a LOESS smooth of variable y as a function of x.
c
c             Usage: call loess(fx,fy,x,y,nx)

c  Here, fx consists of OONP (global, default = 100) points evenly spaced from the minimum x to the maximum x, and fy consists of the
c  fitted values at those points.  Other globals are OOF (tension parameter, default = 0.5), OOSHOW and OOFREQ, the latter two
c  controlling whether and how frequently the procs report progress. A final global OOROBUST determines if robust smoothing is to be
c  used(default = 1).  The variable nx is the number of points in the vector x.
c
c  Local smoothing is linear with neighborhood weights determined by the bi-squarefunction.
c
      implicit integer (a-z)
      parameter (idim=200)
      real yrf,yrl
      double precision x(*),y(*),xs(*),allfit(*),minc,maxc,oof,oof2,rhos(idim),alleps(idim),tmpx(idim),fitted
      double precision medabs,tmp,LCNT(IDIM),YRS(IDIM),YRNOW,  MIN,MAXI
      common /loes/oof,oonp,oofreq,ooshow,oorbst

      oonp=100
      oof=oof2
      oorbst=0

      closef=max(int(nx*oof),1)
      iyrs=ifix(yrl-yrf+1.0)
      do i=1,nx
        rhos(i)=1.0d0
      end do

      if(oorbst.eq.1)then   !  oorbst set to zero, so skip this
        do i=1,nx
          alleps(i)=y(i)-fitted(x,y,rhos,closef,x(i),nx)
        end do
        tmp=medabs(alleps,nx)*6.0d0
        do i=1,nx
          tmpx(i)=alleps(i)/tmp
        end do
        call bisq(rhos,tmpx,nx)
        do i=1,nx
          alleps(i)=y(i)-fitted(x,y,rhos,closef,x(i),nx)
        end do
        tmp=medabs(alleps,nx)*6.0d0
        do i=1,nx
          tmpx(i)=alleps(i)/tmp
        end do
        call bisq(rhos,tmpx,nx)
      endif
      MIN=X(1)
      MAXI=X(NX)
      call xspac(xs,minc(x,nx),maxc(x,nx),oonp)
      DO I=1,iyrs
        IF(YRS(I).ge.(MIN).and.YRS(I).le.(MAXI)) then
          YRNOW=(yrs(i))  !     change made 29 Jan 1996 to make numbering consistent    old= YRNOW=FLOAT(I)
          LCNT(i)=fitted(x,y,rhos,closef,YRNOW,nx)
        end if
      end do
      do i=1,oonp
        if(ooshow.eq.1) write(*,'(1h+,i6)')i
        allfit(i)=fitted(x,y,rhos,closef,xs(i),oonp)
      end do
      return
      end

      function minc(x,n)  !      Find the minimum value of vector x
      double precision x(*),minc
      m=1
      do i=1,n
        if(x(i).lt.x(m))m=i
      end do
      minc=x(m)
      return
      end

      function sumc(x,n) !     Find the sum of vector x
      double precision x(*),sumc
      sumc=0
      do i=1,n
        sumc=sumc+x(i)
      end do
      return
      end

      function maxc(x,n)  !     Find the maximum value of vector x
      double precision x(*),maxc
      m=1
      do i=1,n
        if(x(i).gt.x(m))m=i
      end do
      maxc=x(m)
      return
      end

      subroutine xspac(rslt,a,b,n)
      double precision rslt(*),a,b,xinc
      xinc=1.0d0/(dble(float(n+1)))*(b-a)
      rslt(1)=a+xinc
      do i=2,n
        rslt(i)=rslt(i-1)+xinc
      end do
      return
      end

      function medabs(x,n)  !   Find the median of the absolute values of vector x
      parameter (idim=200)
      double precision x(*),medabs,ax(idim),tmp
      do i=1,n
        ax(i)=abs(x(i))
      end do
      do i=1,n-1
        k=i
        do j=i+1,n
          if(ax(j).lt.ax(k))k=j
        end do
        tmp=ax(i)
        ax(i)=ax(k)
        ax(k)=tmp
      end do
      medabs=ax(n/2+1)
      if(iand(n,1).eq.0)medabs=(ax(n/2)+ax(n/2+1))/2.0d0
      return
      end

      function fitted(x,y,rhos,closef,t,n)
      parameter (idim=200)
      integer closef
      double precision x(idim),y(idim),rhos(idim),t,lamda,tlamda,fitted
      double precision v(idim),tmp(idim),a,b,c,d,e,b0,b1
      tlamda=lamda(x,closef,t,n)
      do i=1,n
        tmp(i)=(x(i)-t)/tlamda
      end do
      call bisq(v,tmp,n)
      a=0.0d0
      b=0.0d0
      c=0.0d0
      d=0.0d0
      e=0.0d0
      do i=1,n
        v(i)=(v(i)+1.0d-50)*rhos(i)
        a=a+v(i)*y(i)
        b=b+v(i)
        c=c+v(i)*x(i)
        d=d+v(i)*x(i)*y(i)
        e=e+v(i)*x(i)*x(i)
        continue
      end do
      b1=(b*d-c*a)/(b*e-c*c)
      b0=(a-c*b1)/b
      fitted=b0+b1*t
      return
      end

      subroutine bisq(rslt,x,n)  !    Compute (1-x^2)^2 for abs(x(i)<1
      double precision rslt(*),x(*)
      do i=1,n
        rslt(i)=0
        if(abs(x(i)).lt.1.0)rslt(i)=(1.0-x(i)*x(i))*(1.0-x(i)*x(i))
      end do
      return
      end

      function lamda(x,closef,t,n)
      parameter (idim=200)
      integer closef
      double precision x(*),t,lamda,ax(idim),tmp
      do i=1,n
         ax(i)=abs(x(i)-t)
      end do
      do i=1,n-1
        k=i
        do j=i+1,n
          if(ax(j).lt.ax(k))k=j
        end do
        tmp=ax(i)
        ax(i)=ax(k)
        ax(k)=tmp
      end do
      lamda=ax(closef)
      return
      end
